/* -*- c -*-
 *
 * Author:      James Brister <brister@vix.com> -- berkeley-unix --
 * Start Date:  Thu Dec 28 17:29:05 1995
 * Project:     INN (innfeed)
 * File:        host.c
 * RCSId:       $Id: host.c,v 1.13 1996/04/13 02:51:33 brister Exp $
 *
 * Copyright:   Copyright (c) 1996 by Internet Software Consortium
 *
 *              Permission to use, copy, modify, and distribute this
 *              software for any purpose with or without fee is hereby
 *              granted, provided that the above copyright notice and this
 *              permission notice appear in all copies.
 *
 *              THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE
 *              CONSORTIUM DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *              SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 *              MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET
 *              SOFTWARE CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT,
 *              INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *              WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 *              WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *              TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
 *              USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Description: Implementation of the Host class.
 * 
 */

#if ! defined (lint)
static const char *rcsid = "$Id: host.c,v 1.13 1996/04/13 02:51:33 brister Exp $" ;
static void use_rcsid (const char *rid) {   /* Never called */
  use_rcsid (rcsid) ; use_rcsid (rid) ;
}
#endif


#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/param.h>
#include <stdio.h>

#include "host.h"
#include "tape.h"
#include "connection.h"
#include "article.h"
#include "endpoint.h"
#include "innlistener.h"
#include "msgs.h"


/* the host keeps a couple lists of these */
typedef struct proc_q_elem 
{
    Article article ;
    struct proc_q_elem *next ;
    struct proc_q_elem *prev ;
} *ProcQElem ;

struct host_s 
{
    InnListener listener ;      /* who created me. */
    char *peerName ;            /* name INN calls the remote */
    char *ipName ;              /* IP name/address from config file. */

    Connection *connections ;   /* NULL-terminated list of all connections */
    bool *cxnActive ;           /* true if the corresponding cxn is active */
    u_int maxConnections ;      /* max # of connections to set up */
    u_int activeCxns ;          /* number of connections currently active */
    u_int initialCxns ;         /* number of connections to create at start */
    u_int currentCxn ;          /* index into the connection list where the
                                   next article should be passed */

    bool remoteStreams ;        /* true if remote supports streaming */
    
    u_int maxChecks ;           /* max number of CHECK commands to send */
    u_int articleTimeout ;      /* max time to wait for new article */
    u_int responseTimeout ;     /* max time to wait for response from remote */
    u_short port ;              /* port the remote listens on */

    ProcQElem queued ;          /* articles done nothing with yet. */
    ProcQElem queuedTail ;

    ProcQElem processed ;       /* articles given to a Connection */
    ProcQElem processedTail ;
    
    TimeoutId statsId ;         /* timeout id for stats logging. */

    Tape myTape ;
    
    bool backedUp ;             /* set to true when all cxns are full */
    u_int backlog ;             /* number of arts in `queued' queue */
    u_int backlogSpooled ;      /* count of arts given to tape */
    
    bool loggedBacklog ;        /* true if we already logged the fact */
    bool notifiedChangedRemBlckd ; /* true if we logged a new response 400 */

    u_int artsOffered ;         /* # of articles we offered to remote. */
    u_int artsAccepted ;        /* # of articles succesfully transferred */
    u_int artsNotWanted ;       /* # of articles remote already had */
    u_int artsRejected ;        /* # of articles remote rejected */
    u_int artsDeferred ;        /* # of articles remote asked us to retry */
    u_int artsMissing ;         /* # of articles whose file was missing. */
    u_int artsFromTape ;        /* # of articles we pulled off tape */

    time_t connectTime ;        /* the time the first connection was fully
                                   set up (MODE STREAM and everything
                                   else). */
    time_t spoolTime ;          /* the time the Host had to revert to
                                   spooling articles to tape. */
    time_t lastSpoolTime ;      /* the time the last time the Host had to
                                   revert to spooling articles to tape. */

    Host next ;                 /* for global list of hosts. */
};





  /*
   * Host object private methods.
   */
static void articleGone (Host host, Connection cxn, Article article) ;
static void hostStopSpooling (Host host) ;
static void hostStartSpooling (Host host) ;
static void hostLogStats (Host host, bool final) ;
static void hostStatsTimeoutCbk (TimeoutId tid, void *data) ;
static void backlogToTape (Host host) ;
static void queuesToTape (Host host) ;
static bool amClosing (Host host) ;


  /* article queue management functions */
static Article remHead (ProcQElem *head, ProcQElem *tail) ;
static void queueArticle (Article article, ProcQElem *head, ProcQElem *tail) ;
static bool remArticle (Article article, ProcQElem *head, ProcQElem *tail) ;





  /*
   * Host class data
   */

  /* if true then when a Host logs its stats, it has all its connections
     log theirs too. */
static bool logConnectionStats = false ;

  /* The frequency in seconds with which a Host will log its stats. */
static u_int statsPeriod = STATS_PERIOD ;

static Host gHostList = NULL ;

static u_int gHostCount = 0 ;

  /*******************************************************************/
  /*                  PUBLIC FUNCTIONS                               */
  /*******************************************************************/



  /* Create a new Host object. Called by the InnListener. */

Host newHost (InnListener listener,
              const char *name, 
              const char *ipName,
              u_int artTimeout, 
              u_int respTimeout,
              u_int initialCxns,
              u_int maxCxns,
              u_int maxCheck,
              u_short portNum,
              u_int closePeriod,
              double lowPassHigh,
              double lowPassLow) 
{
  u_int i ;
  Host nh ; 

  ASSERT (maxCxns > 0 && maxCxns < MAX_CONNECTION_COUNT) ;
  ASSERT (maxCheck > 0) ;

  nh =  CALLOC (struct host_s, 1) ;
  ASSERT (nh != NULL) ;

  nh->listener = listener ;
  nh->peerName = strdup (name) ;
  nh->ipName = strdup (ipName) ;
  
  nh->connections = CALLOC (Connection, maxCxns + 1) ;
  ASSERT (nh->connections != NULL) ;

  nh->cxnActive = CALLOC (bool, maxCxns) ;
  ASSERT (nh->cxnActive != NULL) ;

  nh->maxConnections = maxCxns ;
  nh->activeCxns = 0 ;
  nh->initialCxns = initialCxns ;
  nh->currentCxn = 0 ;

  nh->maxChecks = maxCheck ;
  nh->articleTimeout = artTimeout ;
  nh->responseTimeout = respTimeout ;
  nh->port = portNum ;

  nh->queued = NULL ;
  nh->queuedTail = NULL ;

  nh->processed = NULL ;
  nh->processedTail = NULL ;
  
  nh->statsId = 0 ;

    /* Create the Host's tape. If we're only working off backup (listener is
       a dummy InnListener), then any articles we spool to tape don't get
       read back in. */
  nh->myTape = newTape (name,listenerIsDummy (listener)) ;
  if (nh->myTape == NULL)
    {                           /* tape couldn't be locked, probably */
      FREE (nh) ;
      return NULL ;
    }

  nh->backedUp = false ;
  nh->backlog = 0 ;
  nh->backlogSpooled = 0 ;

  nh->loggedBacklog = false ;
  nh->notifiedChangedRemBlckd = false ;

  nh->artsOffered = 0 ;
  nh->artsAccepted = 0 ;
  nh->artsNotWanted = 0 ;
  nh->artsRejected = 0 ;
  nh->artsDeferred = 0 ;
  nh->artsMissing = 0 ;
  
  nh->connectTime = 0 ;
  
  nh->spoolTime = 0 ;

    /* Create all the connections, but only the initial ones connect
       immediately */
  for (i = 0 ; i < maxCxns ; i++)
    {
      nh->cxnActive [i] = false ;
      nh->connections [i] = newConnection (nh,
                                           i,
                                           nh->ipName,
                                           nh->articleTimeout,
                                           nh->port,
                                           nh->responseTimeout,
                                           closePeriod,
                                           lowPassHigh,
                                           lowPassLow) ;
      if (i < initialCxns)
        cxnConnect (nh->connections [i]) ;
    }

  nh->connections [maxCxns] = NULL ;

  nh->next = gHostList ;
  gHostList = nh ;
  gHostCount++ ;
  
  return nh ;
}


void gPrintHostInfo (FILE *fp, u_int indentAmt)
{
  Host h ;
  char indent [INDENT_BUFFER_SIZE] ;
  u_int i ;
  
  for (i = 0 ; i < MIN(INDENT_BUFFER_SIZE - 1,indentAmt) ; i++)
    indent [i] = ' ' ;
  indent [i] = '\0' ;
  
  fprintf (fp,"%sGlobal Host list : (count %d) {\n",indent,gHostCount) ;
  
  for (h = gHostList ; h != NULL ; h = h->next)
    printHostInfo (h,fp,indentAmt + INDENT_INCR) ;
  
  fprintf (fp,"%s}\n",indent) ;
}


void printHostInfo (Host host, FILE *fp, u_int indentAmt)
{
  char indent [INDENT_BUFFER_SIZE] ;
  u_int i ;
  Connection *cxn ;
  ProcQElem qe ;
  
  for (i = 0 ; i < MIN(INDENT_BUFFER_SIZE - 1,indentAmt) ; i++)
    indent [i] = ' ' ;
  indent [i] = '\0' ;

  fprintf (fp,"%sHost : %p {\n",indent,host) ;

  if (host == NULL)
    {
      fprintf (fp,"%s}\n",indent) ;
      return ;
    }
  
  fprintf (fp,"%s    peer-name : %s\n",indent,host->peerName) ;
  fprintf (fp,"%s    ip-name : %s\n",indent,host->ipName) ;
  fprintf (fp,"%s    max-connections : %d\n",indent,host->maxConnections) ;
  fprintf (fp,"%s    active-connections : %d\n",indent,host->activeCxns) ;
  fprintf (fp,"%s    current-connection : %d\n",indent,host->currentCxn) ;
  fprintf (fp,"%s    remote-streams : %s\n",indent,
           boolToString (host->remoteStreams)) ;
  fprintf (fp,"%s    max-checks : %d\n",indent,host->maxChecks) ;
  fprintf (fp,"%s    article-timeout : %d\n",indent,host->articleTimeout) ;
  fprintf (fp,"%s    resposne-timeout : %d\n",indent,host->responseTimeout) ;
  fprintf (fp,"%s    port : %d\n",indent,host->port) ;
  fprintf (fp,"%s    statistics-id : %d\n",indent,host->statsId) ;
  fprintf (fp,"%s    backed-up : %s\n",indent,boolToString (host->backedUp));
  fprintf (fp,"%s    backlog : %d\n",indent,host->backlog) ;
  fprintf (fp,"%s    backlogSpooled : %d\n",indent,host->backlogSpooled) ;
  fprintf (fp,"%s    logged-backlog : %s\n",indent,
           boolToString (host->loggedBacklog)) ;
  fprintf (fp,"%s    streaming-type changed : %s\n",indent,
           boolToString (host->notifiedChangedRemBlckd)) ;
  fprintf (fp,"%s    articles offered : %d\n",indent,host->artsOffered) ;
  fprintf (fp,"%s    articles accepted : %d\n",indent,host->artsAccepted) ;
  fprintf (fp,"%s    articles not wanted : %d\n",indent,
           host->artsNotWanted) ;
  fprintf (fp,"%s    articles rejected : %d\n",indent,host->artsRejected);
  fprintf (fp,"%s    articles deferred : %d\n",indent,host->artsDeferred) ;
  fprintf (fp,"%s    articles missing : %d\n",indent,host->artsMissing) ;

  fprintf (fp,"%s    connectTime : %s",indent,ctime (&host->connectTime));
  fprintf (fp,"%s    spoolTime : %s",indent,ctime (&host->spoolTime)) ;
  fprintf (fp,"%s    last-spool-time : %s",indent,
           ctime (&host->lastSpoolTime)) ;
  
#if 0
  fprintf (fp,"%s    tape {\n",indent) ;
  printTapeInfo (host->myTape,fp,indentAmt + INDENT_INCR) ;
  fprintf (fp,"%s    }\n",indent) ;
#else
  fprintf (fp,"%s    tape : %p\n",indent,host->myTape) ;
#endif
  
  fprintf (fp,"%s    QUEUED articles {\n",indent) ;
  for (qe = host->queued ; qe != NULL ; qe = qe->next)
    {
#if 0
      printArticleInfo (qe->article,fp,indentAmt + INDENT_INCR) ;
#else
      fprintf (fp,"%s    %p\n",indent,qe->article) ;
#endif
    }
  
  fprintf (fp,"%s    }\n",indent) ;
  
  fprintf (fp,"%s    IN PROCESS articles {\n",indent) ;
  for (qe = host->processed ; qe != NULL ; qe = qe->next)
    {
#if 0
      printArticleInfo (qe->article,fp,indentAmt + INDENT_INCR) ;
#else
      fprintf (fp,"%s    %p\n",indent,qe->article) ;
#endif
    }
  
  fprintf (fp,"%s    }\n",indent) ;

  
  
  fprintf (fp,"%s    Connections {\n",indent) ;
  for (cxn = host->connections ; cxn != NULL && *cxn != NULL ; cxn++)
    {
#if 0
    printCxnInfo (*cxn,fp,indentAmt + INDENT_INCR) ;
#else
    fprintf (fp,"%s        %p\n",indent,*cxn) ;
#endif
    }
  fprintf (fp,"%s    }\n",indent) ;

  fprintf (fp,"%s    Active Connections {\n%s        ",indent,indent) ;
  for (i = 0 ; i < host->maxConnections ; i++)
    if (host->cxnActive[i])
      fprintf (fp," [%d:%p]",i,host->connections[i]) ;
  fprintf (fp,"\n%s    }\n",indent) ;

  fprintf (fp,"%s}\n",indent) ;
}



  /* close down all the connections of the Host. All articles that are in
     processes are still pushed out and then a QUIT is issued. The Host
     will also spool all inprocess articles to tape incase the process is
     about to be killed (they'll be refused next time around). When all
     Connections report that they're gone, then the Host will delete itself. */
void hostClose (Host host)
{
  u_int i ;

  dprintf (1,"Closing host %s\n",host->peerName) ;
  
  queuesToTape (host) ;
  delTape (host->myTape) ;
  host->myTape = NULL ;
  
  hostLogStats (host,true) ;

  removeTimeout (host->statsId) ;
  host->statsId = 0 ;
  
  host->connectTime = 0 ;

  for (i = 0 ; i < host->maxConnections ; i++)
    cxnClose (host->connections [i]) ;
}



  /* have the Host transmit the Article if possible. */
void hostSendArticle (Host host, Article article)
{
  if (host->spoolTime != 0)
    {
      host->backlogSpooled++ ;
      tapeTakeArticle (host->myTape, article) ;
    }
  else if (host->activeCxns == 0)
    {
      dprintf (1,"No active connections to %s for article\n",host->peerName) ;

        /* stick on the queue of articles we've done nothing with yet. */
      queueArticle (article,&host->queued,&host->queuedTail) ;
      host->backlog++ ;
      if (host->backlog > HOST_HIGHWATER)
        hostStartSpooling (host) ;
    }
  else
    {
      u_int idx = host->currentCxn ;
      Article extraRef ;
      bool taken = false ;
      
      extraRef = artTakeRef (article) ; /* the referrence we give away */
      
        /* stick on the queue of articles we've handed off. */
      queueArticle (article,&host->processed,&host->processedTail) ;

        /* first we try to give it to one of our active connections */
      do
        {
          if (host->cxnActive [idx])
            taken = cxnTakeArticle (host->connections [idx],extraRef) ;
          
          idx = (idx + 1) % host->maxConnections ;
        }
      while (!taken && idx != host->currentCxn) ;

      if ( taken )
        host->currentCxn = idx ;
      else
        {
            /* Wasn't taken so try to give it to one of the inactive
               connections. This will cause the connection to try and
               connect up (if it isn't in the middle of that already). */

          for (idx = 0 ; idx < host->maxConnections ; idx++)
            if (!host->cxnActive [idx])
              {
                if (cxnTakeArticle (host->connections [idx], extraRef))
                  break ;
                else
                  dprintf (1,"%s Inactive connection %d refused an article\n",
                           host->peerName,idx) ;
              }

          if (idx == host->maxConnections) /* it did not get taken */
            {
              dprintf (1, "Couldn't give the article to a connection\n") ;
              
              delArticle (extraRef) ;
                  
              remArticle (article,&host->processed,&host->processedTail) ;
              queueArticle (article,&host->queued,&host->queuedTail) ;
              
              host->backlog++ ;
              if (host->backlog > HOST_HIGHWATER)
                backlogToTape (host) ;
            }
        }
    }
}





  /* called by the Host's connection when the remote is refusing postings
     from us becasue we're not allowed (banner code 400). */
void hostCxnBlocked (Host host, Connection cxn, char *reason)
{
#ifndef NDEBUG
  {
    u_int i ;
    
    for (i = 0 ; i < host->maxConnections ; i++)
      if (host->connections [i] == cxn)
        ASSERT (host->cxnActive [i] == false) ;
  }
#endif

  if (host->activeCxns == 0 && host->spoolTime == 0)
    syslog (LOG_NOTICE,REMOTE_BLOCKED, host->peerName, reason) ;
  else if (host->activeCxns > 0 && !host->notifiedChangedRemBlckd)
    {
      syslog (LOG_NOTICE,CHANGED_REMOTE_BLOCKED, host->peerName,reason) ;
      host->notifiedChangedRemBlckd = true ;
    }
}



  /* Called by the Connection when it gets a response back to the MODE
     STREAM command. It's now that we consider the connection usable. */
void hostRemoteStreams (Host host, Connection cxn, bool doesStreaming)
{
  u_int i ;

    /* we may have told the connection to quit while it was in the middle
       of connecting */
  if (amClosing (host))
    return ;
  
  if (host->connectTime == 0)   /* first connection for this cycle. */
    {
      host->connectTime = time (NULL) ;

      if (doesStreaming)
        syslog (LOG_NOTICE, REMOTE_DOES_STREAMING, host->peerName);
      else
        syslog (LOG_NOTICE, REMOTE_NO_STREAMING, host->peerName);

        /* set up the callback for statistics logging. */
      ASSERT (host->statsId == 0) ;
      host->statsId = prepareSleep (hostStatsTimeoutCbk, statsPeriod, host) ;

      host->remoteStreams = doesStreaming ;

      if (host->spoolTime != 0)
        hostStopSpooling (host) ;
    }
  else if (host->remoteStreams != doesStreaming)
    syslog (LOG_NOTICE,STREAMING_CHANGE,host->peerName) ;

  for (i = 0 ; i < host->maxConnections ; i++)
    if (host->connections [i] == cxn)
      {
        host->cxnActive [i] = true ;
        break ;
      }

  ASSERT (i != host->maxConnections) ;
  
  host->activeCxns++ ;
}


  /* Called by the connection when it is no longer connected to the
     remote. Perhaps due to getting a code 400 to an IHAVE, or due to a
     periodic close. */
void hostCxnDead (Host host, Connection cxn)
{
  u_int i ;
    
  for (i = 0 ; i < host->maxConnections ; i++)
    if (host->connections [i] == cxn)
      {
        if (host->cxnActive [i]) /* won't be active if got 400 on banner */
          {
            host->cxnActive [i] = false ;
            host->activeCxns-- ;

            if (host->activeCxns == 0)
              hostLogStats (host,true) ;
          }
      }
}


  /* Called by the Connection when it is about to delete itself. */
void hostCxnGone (Host host, Connection cxn)
{
  u_int i ;
  bool oneThere = false ;

    /* forget abou the Connection and see if we are still holding any live
       connections still. */
  for (i = 0 ; i < host->maxConnections ; i++)
    if (host->connections [i] == cxn)
      host->connections [i] = NULL ;
    else if (host->connections [i] != NULL)
      oneThere = true ;

  if ( !oneThere )
    {
      InnListener l = host->listener ;
      
      delHost (host) ;
      listenerHostGone (l, host) ;
    }
}






  /* The connections has offered an article to the remote. */
void hostArticleOffered (Host host, Connection cxn) 
{
  (void) cxn ;                  /* keep lint happy. */
  
  host->artsOffered++ ;
}


  /* Article was succesfully transferred. */
void hostArticleAccepted (Host host, Connection cxn, Article article)
{
  const char *filename = artFileName (article) ;
  const char *msgid = artMsgId (article) ;

  dprintf (5,"Article %s (%s) was transferred\n", msgid, filename) ;
  
  host->artsAccepted++ ;

    /* host has two references to the article here... the parameter `article'
       and the queue */

  delArticle (article) ;        /* drop the parameter reference */

  if (!amClosing (host))
    articleGone (host,cxn,article) ; /* and the one in the queue */
}


  /* remote said no thanks. */
void hostArticleNotWanted (Host host, Connection cxn, Article article)
{
  const char *filename = artFileName (article) ;
  const char *msgid = artMsgId (article) ;

  dprintf (5,"Article %s (%s) was not wanted\n", msgid, filename) ;
  
  host->artsNotWanted++ ;
  
    /* host has two references to the article here... `article' and the
       queue */

  delArticle (article) ;        /* drop the `article' reference */
  
  if (!amClosing (host)) 
    articleGone (host,cxn,article) ; /* and the one in the queue */
}


  /* remote rejected the article after it was was transferred */
void hostArticleRejected (Host host, Connection cxn, Article article) 
{
  const char *filename = artFileName (article) ;
  const char *msgid = artMsgId (article) ;

  dprintf (5,"Article %s (%s) was rejected\n", msgid, filename) ;
  
  host->artsRejected++ ;

  /* host has two references to the article here... `article' and the queue */

  delArticle (article) ;        /* drop the `article' reference */

  if (!amClosing (host))
    articleGone (host,cxn,article) ;
}


  /* The remote wants us to retry the article later. */
void hostArticleDeferred (Host host, Connection cxn, Article article) 
{
  host->artsDeferred++ ;

  host->backlogSpooled++ ;

  if (!amClosing (host))
    {
      tapeTakeArticle (host->myTape,article) ;
      articleGone (host,cxn,article) ; /* drop from the queue */
    }
}


  /* The disk file for the article is no longer valid */
void hostArticleIsMissing (Host host, Connection cxn, Article article)
{
  const char *filename = artFileName (article) ;
  const char *msgid = artMsgId (article) ;

  dprintf (5, "%s article is missing %s %s\n", host->peerName, msgid, filename) ;
    
  host->artsMissing++ ;

    /* host has two references to the article here... `article' and the
       queue */

  delArticle (article) ;        /* drop the `article' reference */

  if (!amClosing (host))
    articleGone (host,cxn,article) ; /* and the one in the queue */
}


  /* The Connection wants something to do. This is called by the Connection
     after it has transferred an article. This is what keeps the pipes full
     of data off the tapes if the input from inn is idle. */
bool hostGimmeArticle (Host host, Connection cxn)
{
  Article article = NULL ;
  bool gaveSomething = false ;
  size_t amtToGive = cxnQueueSpace (cxn) ; /* may be more than one */

  if (amClosing (host))
    {
      dprintf (5,"%s no article to give due to closing\n",host->peerName) ;

      return false ;
    }

  if (amtToGive == 0)
    dprintf (5,"%s Queue space is zero....\n",host->peerName) ;
  
  while (amtToGive > 0)
    {
      bool tookIt ;
      
      if ((article = remHead (&host->queued,&host->queuedTail)) != NULL)
        {
          host->backlog-- ;
          tookIt = cxnQueueArticle (cxn,artTakeRef (article)) ;

          ASSERT (tookIt == true) ;

          queueArticle (article,&host->processed,&host->processedTail) ;
          amtToGive-- ;

          gaveSomething = true ;
        }
      else if ((article = getArticle (host->myTape)) != NULL) 
        {                       /* go to the tapes */
          tookIt = cxnQueueArticle (cxn,artTakeRef (article)) ;

          ASSERT (tookIt == true) ;

          host->artsFromTape++ ;
          queueArticle (article,&host->processed,&host->processedTail) ;
          amtToGive-- ;

          gaveSomething = true ;
        }
      else
        {
            /* we had nothing left to give... */
          
          if (host->processed == NULL) /* and if nothing outstanding... */
            listenerHostIsIdle (host->listener,host) ; /* tell our owner */
  
          amtToGive = 0 ;
        }
    }

  return gaveSomething ;
}





/* get the name that INN uses for this host */
const char *hostPeerName (Host host)
{
  ASSERT (host != NULL) ;
    
  return host->peerName ;
}


u_int hostMaxChecks (Host host)
{
  return host->maxChecks ;
}




  /**********************************************************************/
  /**                       CLASS FUNCTIONS                            **/
  /**********************************************************************/


void hostLogConnectionStats (bool val)
{
  logConnectionStats = val ;
}

bool hostLogConnectionStatsP (void)
{
  return logConnectionStats ;
}

void hostSetStatsPeriod (u_int period) 
{
  statsPeriod = period ;
}





  /**********************************************************************/
  /**                      PRIVATE FUNCTIONS                           **/
  /**********************************************************************/


  /* fully delete and clean up the Host object. */
void delHost (Host host)
{
  Host h,q ;

  for (h = gHostList, q = NULL ; h != NULL ; q = h, h = h->next)
    if (h == host)
      {
        if (gHostList == h)
          gHostList = gHostList->next ;
        else
          q->next = h->next ;
        break ;
      }

  ASSERT (h != NULL) ;
        
  delTape (host->myTape) ;
  
  FREE (host->connections) ;
  FREE (host->cxnActive) ;
  FREE (host->peerName) ;
  FREE (host->ipName) ;

  FREE (host) ;
}


  /* the article can be dropped from the process queue and the connection
     can take a new article if there are any to be had. */
static void articleGone (Host host, Connection cxn, Article article)
{
  if ( !remArticle (article,&host->processed,&host->processedTail) )
    die ("remArticle in hostArticleDeferred failed") ;

  delArticle (article) ;

  hostGimmeArticle (host,cxn) ; /* may not give anything over */
}


  /* One of the Connections for this Host has reestablished itself, so stop
     spooling article info to disk. */
static void hostStopSpooling (Host host)
{
  ASSERT (host->spoolTime != 0) ;
  
  tapeClose (host->myTape) ;
  host->spoolTime = 0 ;
}


  /* No connections are active and we're getting response 201 or 400 (or
     some such) so that we need to start spooling article info to disk. */
static void hostStartSpooling (Host host)
{
  ASSERT (host->spoolTime == 0) ;

  hostLogStats (host,true) ;
  
  host->spoolTime = time (NULL) ;

    /* don't want to log too frequently */
  if (SPOOL_LOG_PERIOD > 0 &&
      (host->spoolTime - host->lastSpoolTime) > SPOOL_LOG_PERIOD)
    syslog (LOG_NOTICE,SPOOLING,host->peerName) ;

  host->lastSpoolTime = host->spoolTime ;
  host->connectTime = 0 ;

  host->notifiedChangedRemBlckd = false ;

  removeTimeout (host->statsId) ;
  host->statsId = prepareSleep (hostStatsTimeoutCbk, statsPeriod, host) ;

  /* while we're spooling everything goes straight to disk */
  tapeOpen (host->myTape) ;
}


  /* Time to log the statistics for the Host. If FINAL is true then the
     counters will be reset. */
static void hostLogStats (Host host, bool final)
{
  time_t now = time (NULL) ;
  time_t startPeriod ;

  if (host->spoolTime == 0 && host->connectTime == 0)
    return ;                    /* host has never connected */

  startPeriod = (host->spoolTime != 0 ? host->spoolTime : host->connectTime) ;

  if (now - startPeriod >= STATS_RESET_PERIOD)
    final = true ;
  
  if (host->spoolTime != 0)
    syslog (LOG_NOTICE, HOST_SPOOL_STATS, host->peerName,
            (final ? "final" : "checkpoint"),
            (long) (now - host->spoolTime), host->backlogSpooled) ;
  else
    syslog (LOG_NOTICE, HOST_STATS_MSG, host->peerName, 
            (final ? "final" : "checkpoint"),
            (long) (now - host->connectTime),
            host->artsOffered, host->artsAccepted,
            host->artsNotWanted, host->artsRejected,
            host->artsMissing,host->backlogSpooled) ;

  if (logConnectionStats) 
    {
      u_int i ;
      
      for (i = 0 ; i < host->activeCxns ; i++)
        cxnLogStats (host->connections [i],final) ;
    }

  /* one 'spooling backlog' message per stats logging period */
  host->loggedBacklog = false ;

  if (final)
    {
      host->artsOffered = 0 ;
      host->artsAccepted = 0 ;
      host->artsNotWanted = 0 ;
      host->artsRejected = 0 ;
      host->artsDeferred = 0 ;
      host->artsMissing = 0 ;
      host->artsFromTape = 0 ;
      host->backlogSpooled = 0 ;

      host->connectTime = time (NULL) ; /* in of case STATS_RESET_PERIOD */
    }

#if 0
    /* turn this section on to get a snapshot at each log period. */
  if (gPrintInfo != NULL)
    gPrintInfo () ;
#endif
}


  /* The callback function for the statistics timer to call. */
static void hostStatsTimeoutCbk (TimeoutId tid, void *data)
{
  Host host = (Host) data ;

  (void) tid ;                  /* keep lint happy */

  ASSERT (tid == host->statsId) ;
  
  hostLogStats (host, false) ;
  
  host->statsId = prepareSleep (hostStatsTimeoutCbk, statsPeriod, host) ;
}


  /* the host has too many unprocessed articles so we send them to the tape. */
static void backlogToTape (Host host)
{
  Article article ;

  if (!host->loggedBacklog)
    {
      syslog (LOG_NOTICE,BACKLOG_TO_TAPE,host->peerName /* ,host->backlog */) ;
      host->loggedBacklog = true ;
    }
  
  while ((article = remHead (&host->queued,&host->queuedTail)) != NULL)
    {
      host->backlogSpooled++ ;
      tapeTakeArticle (host->myTape,article) ;
    }

  host->backlog = 0 ;
}



static bool amClosing (Host host)
{
  return (host->myTape == NULL ? true : false) ;
}



  /* flush all queued articles all the way out to disk. */
static void queuesToTape (Host host)
{
  Article art ;
  
  tapeOpen (host->myTape) ;

  while ((art = remHead (&host->processed,&host->processedTail)) != NULL)
    {
      host->backlogSpooled++ ;
      tapeTakeArticle (host->myTape,art) ;
    }
  
  while ((art = remHead (&host->queued,&host->queuedTail)) != NULL)
    {
      host->backlog-- ;
      host->backlogSpooled++ ;
      tapeTakeArticle (host->myTape,art) ;
    }

  tapeClose (host->myTape) ;
}



  /* Add an article to the given queue. */
static void queueArticle (Article article, ProcQElem *head, ProcQElem *tail)
{
  ProcQElem elem ;
  
  elem = ALLOC (struct proc_q_elem, 1) ;
  ASSERT (elem != NULL) ;

  elem->article = article ;
  elem->next = NULL ;
  elem->prev = *tail ;
  if (*tail != NULL)
    (*tail)->next = elem ;
  else
    *head = elem ;
  *tail = elem ;  
}


  /* remove the article from the queue */
static bool remArticle (Article article, ProcQElem *head, ProcQElem *tail)
{
  ProcQElem elem ;

  ASSERT (head != NULL) ;
  ASSERT (tail != NULL) ;

  /* we go backwards down the list--probably faster */
  elem = *tail ;
  while (elem != NULL && elem->article != article)
    elem = elem->prev ;

  if (elem != NULL)
    {
      if (elem->prev != NULL)
        elem->prev->next = elem->next ;
      if (elem->next != NULL)
        elem->next->prev = elem->prev ;
      if (*head == elem)
        *head = elem->next ;
      if (*tail == elem)
        *tail = elem->prev ;

      FREE (elem) ;
      
      return true ;
    }
  else
    return false ;
}


  /* remove the article that's at the head of the queue and return
     it. Returns NULL if the queue is empty. */
static Article remHead (ProcQElem *head, ProcQElem *tail)
{
  ProcQElem elem ;
  Article art ;

  ASSERT (head != NULL) ;
  ASSERT (tail != NULL) ;
  ASSERT ((*head == NULL && *tail == NULL) ||
          (*head != NULL && *tail != NULL)) ;

  if (*head == NULL)
    return NULL ;

  elem = *head ;
  art = elem->article ;
  *head = elem->next ;
  if (elem->next != NULL)
    elem->next->prev = NULL ;

  if (*tail == elem)
    *tail = NULL ;

  FREE (elem) ;

  return art ;
}



























  






#if 0
  /* Have all the Connections disconnect from their remotes and then
     reconnect automatically. */
void hostFlush (Host host)
{
  u_int i ;

  dprintf (1,"Flushing host %s\n",host->peerName) ;
  
  host->activeCxns = 0 ;

  hostLogStats (host,true) ;
  
  for (i = 0 ; i < host->maxConnections ; i++)
    if (host->cxnActive [i])
        /* XXX what about a connection that's no active yet, but is being
           set up? */
      cxnFlush (host->connections [i]) ;
}
#endif





#if 0


static void initializeConnection (Host host) ;
static void delConnectionPvt (Host host, Connection cxn) ;
static void hostStatsTimeoutCbk (TimeoutId tid, void *data) ;
static void reEstablishCxn (TimeoutId tid, void *data) ;
static void hostLogStats (Host host, bool final) ;
static void hostStartSpooling (Host host) ;
static void articleGone (Host host, Connection cxn, Article article) ;



/* return true if the host has nothing to do. */
bool hostIsIdle (Host host)
{
  bool rval ;

#if 0

  rval = (host->queued == NULL) &&
    (host->processed == NULL) &&
    !tapeHasArticles (host->inputTape) &&
    !tapeHasArticles (host->outputTape) ;

#else

  rval = (host->queued == NULL) &&
    (host->processed == NULL) &&
    !tapeHasArticles (host->myTape) ;
  
#endif

  return rval ;
}

  /* the directory where locks on hosts get put. */
static const char *lockDirName ;


bool lockHost (const char *peerName)
{
  char lockFileName [MAXPATHLEN] ;
  
  if (lockDirName == NULL)
    lockDirName = getTapeDirectory () ;

  sprintf (lockFileName,"%s/LOCK.%s",lockDirName,peerName) ;

  return lockFile (lockFileName) ;
}


void setHostLockDir (const char *dirname)
{
  if (lockDirName != NULL)
    warn ("To late. Host lock directory is already set.\n") ;
  else
    lockDirName = strdup (dirname) ;
}





/* timeout callback function for when the Host's Connections are unable to
   contact the remote. This is how we keep trying. */
static void reEstablishCxn (TimeoutId tid, void *data)
{
  Host host = (Host) data ;

  (void) tid ;                  /* keep lint happy */

  ASSERT (tid == host->reEstabId) ;
  
  initializeConnection (host) ;

  if (host->spooling == true)   /* couldn't make any for some reason */
    {
      host->reEstabId = prepareSleep (reEstablishCxn,
                                      host->reEstabPeriod, host) ;

      host->reEstabPeriod *= 1.5 ;
      if (host->reEstabPeriod > MAX_REESTABLISHMENT_PERIOD)
        host->reEstabPeriod = MAX_REESTABLISHMENT_PERIOD ;
    }
  else
    {
      bool haveCon = false ;
      u_int i ;

      /* on a fast machine connecting locally, the non-blocking
         connect() may have connected anyway */
      for (i = 0 ; i < host->activeCxns ; i++)
        haveCon = haveCon || cxnIsConnected (host->connections [i]) ;

      if (haveCon)
        {
          syslog (LOG_NOTICE,CXN_RESTABISH,host->peerName) ;
          host->reEstabId = 0 ;
        }

      /* we leave the reEstabId alone so we can tell later we were trying
         to reestablish   */
    }
}


/*
 * Called when there are no active connections and we want to set one up.
 * This could be at the creation of the Host, or it could be later when a
 * remote has been unreachable and we've been idly spooling articles to
 * Tape.
 */
static void initializeConnection (Host host)
{
  u_int i ;
  
  ASSERT (host->activeCxns == 0) ;
  
  for (i = 0 ; i < host->initialCxns ; i++)
    {
      host->connections [i] = newConnection (host,
                                             i,
                                             host->ipName,
                                             host->maxChecks,
                                             host->articleTimeout,
                                             host->immediateOpen,
                                             host->port,
                                             host->responseTimeout,
                                             host->closePeriod) ;

      if (host->connections[i] != NULL)
        host->activeCxns++ ;
      else
        break ;
    }

  if (host->activeCxns == 0)
      hostStartSpooling (host) ;
  else
    host->spooling = false ;
}

/* delete the connection from the host */
static void delConnectionPvt (Host host, Connection cxn) 
{
  u_int idx ;

  ASSERT (cxn != NULL) ;
  
  for (idx = 0 ; idx < host->maxConnections ; idx++)
    if (host->connections [idx] == cxn)
      break ;

  ASSERT (idx != host->maxConnections) ;

  delConnection (host->connections [idx]) ;
  host->activeCxns-- ;
  if (host->currentCxn == host->activeCxns && host->activeCxns > 0)
    host->currentCxn-- ;
  
  for ( ; idx < host->maxConnections - 1 ; idx++ )
    host->connections[idx] = host->connections [idx + 1] ;

  host->connections [idx] = NULL ;
}



  /* called by the Connection when it gets a code 201 for the banner
     message */
void hostCxnNoPerm (Host host, Connection cxn)
{
#ifndef NDEBUG
  {
    int i ;
    
    for (i = 0 ; i < host->maxConnections ; i++)
      if (host->connections [i] == cxn)
        ASSERT (host->cxnActive [i] == false) ;
  }
#endif

  if (host->activeCxns == 0 && host->spoolTime == 0)
    {
      syslog (LOG_NOTICE,NO_POST_PERMISSION, host->peerName) ;

      hostStartSpooling (host) ;
    }
  else if (host->activeCxns > 0 && !host->notifiedChangedPostPerm)
    {
      syslog (LOG_NOTICE,CHANGED_POST_PERMISSION, host->peerName) ;
      host->notifiedChangedPostPerm = true ;
    }
}




#endif



